# $Header: /cvsroot/SchemaEditor/SchemaEditor.py,v 1.64 2003/09/05 14:38:53 spuelrich Exp $
from OFS.SimpleItem import SimpleItem

from Persistence import Persistent
from Products.PythonScripts.standard import html_quote
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
import time

from xsdTools import xsdURI, getChildNodesByTagNameNS, minimize_paths, hasAncestorWithNameNS
from XSPrettyPrinter import TheXSPrettyPrinter
import XSTypeRegistry
import XSWebType
import XSCustomTypes
from XSCardinality import XSCardinality
import XSExceptions

import install

class SchemaEditorSession(Persistent):
    def __init__(self, xsd, schema_editor):
        self.type_registry = XSTypeRegistry.XSTypeRegistry()
        self.xsd_modificationtime = None
        self.values = {}
        self.init_image_urls(schema_editor)
        return
        
    def init_image_urls(self, schema_editor):
        image_tag_templates = {}
        # use Acquisiotn to get the globals folder
        globals = schema_editor.globals
        for image_name in ('edit', 'add', 'replace', 'delete', 'info'):
            image = getattr(globals, image_name+'.png')
            image_url = '/'.join(image.getPhysicalPath())
            templateformat = '<a title="%%s" href="%%s">'
            templateformat += '<img class="linkimage" src="%s" alt="%%s" height="%d" width="%d"/>'
            templateformat += '</a>'
            template = templateformat % (image_url, image.height, image.width)
            # title, href, id, id, id, alt
            image_tag_templates[image_name] = template

        self.image_tag_templates = image_tag_templates
        return

    def get_image_link(self, url, image_name, alt_text):
        # extend the template by
        # a-title, href, id, id, id, alt
        image_link = self.image_tag_templates[image_name] % (
            alt_text, url, alt_text)
        return image_link

    def check(self, xsd):
        """checks whether the XSD has changed since the last check and
        reinitialises the type registries if necessary."""
        if self.xsd_modificationtime != xsd.bobobase_modification_time():
            self.xsd_modificationtime = xsd.bobobase_modification_time()
            self.type_registry.clear()
            self._initialize_complex_types(xsd)
            self._initialize_simple_types(xsd)

    def _initialize_complex_types(self, xsd):
        typenodes = xsd.documentElement.getElementsByTagNameNS(xsdURI, 'complexType')
        for node in typenodes:
            xstype = XSCustomTypes.XSComplexType(node)
            self.type_registry.register_type(xstype)
        return

    def _initialize_simple_types(self, xsd):
        typenodes = xsd.documentElement.getElementsByTagNameNS(xsdURI, 'simpleType')
        for node in typenodes:
            if hasAncestorWithNameNS(node, 'simpleType', xsdURI):
                # simpleTypes within simpleTypes are handeled by the XSCustomType.XSSimpleType function
                continue
            xstype = XSCustomTypes.XSSimpleType(node)
            self.type_registry.register_type(xstype)

    def show_info(self, REQUEST):
        output = self.type_registry.info()
        output += "\n"
        output += 'xsd_modificationtime = %s\n' % self.xsd_modificationtime
        if self.values:
            for k, v in self.values.items():
                output += "%s = %s\n" % (k,v)
                
        output += "\ninfo from XMLWidgets session\n"
        xs = REQUEST.SESSION['xmlwidgets_service_editor'][REQUEST['xsd_doc_physicalpath']]
        output += "  roots:    %s\n" % str(xs.roots)
        output += "  contexts: %s\n" % str(xs.contexts)
        output += "  document_context: %s\n" % str(xs.document_context)
 
        return output

    def set_edit_node(self, node, xsi_boundary_node=None):
        values = self.values
        values['edit_node'] = node.getNodePath('xsi')
        if xsi_boundary_node is not None:
            values['xsi_boundary_node'] = xsi_boundary_node.getNodePath('element_id')
        else:
            values['xsi_boundary_node'] = None
        self.values = values

    def get_boundary_node(self):
        return self.values.get('xsi_boundary_node', None)

    def clear_edit_node(self):
        values = self.values
        values['edit_node'] = None
        values['xsi_boundary_node'] = None
        self.values = values

    def is_edit_node(self, node, nodepath=None, check_ancestors=0):
        if nodepath is None:
            nodepath = node.getNodePath('xsi')
        edit_nodepath = self.values.get('edit_node', None)
        if edit_nodepath is None:
            return 0
        if not check_ancestors:
            return (edit_nodepath == nodepath)
        else:
            nodepath = nodepath.split(',')
            edit_nodepath = edit_nodepath.split(',')
            if nodepath[:len(edit_nodepath)] == edit_nodepath:
                return 1
            else:
                return 0

class SchemaEditor(SimpleItem):
    """Der SchemaEditor"""

    meta_type = 'Schema Editor'
    setXmlDocsForm = PageTemplateFile("www/setXmlDocsForm", globals())
    index_html = PageTemplateFile('www/SchemaEditor_index_html', globals())
    result_page = PageTemplateFile('www/result_page', globals())
    
    def __init__(self, id):
        self.id = id

    def log(self, text):
        """for debugging from python scripts"""
        print "SchemaEditorLog: %s" % text
        return

    def show_info(self):
        """shows info about the current SchemaEditorSession"""
        session = self.get_session()
        if session is not None:
            return session.show_info(self.REQUEST)
        return ''

    def search_xsd_xsi(self, redirect=1):
        """sucht in aq_chains nach xsd und xsi.
        das am weitesten hinten stehende Parsed XML wird xsd, das vorige xsi. die brigen werden
        ignoriert. /xml/xml/xsi/xsd/schema_editor"""
        xmldocs = []
        chain = self.aq_chain
        chain.reverse()
        for obj in chain:
            if getattr(obj, 'meta_type', '') == 'Parsed XML':
                xmldocs.append(obj)
            if len(xmldocs) >= 2:
                # return als (xsi, xsd)
                return xmldocs[:2]
        if redirect:
            self.REQUEST.RESPONSE.redirect(self.REQUEST['URL1'] + '/setXmlDocsForm')
        return None

    def renderContent(self):
        """
        Sucht nach den XML Dokumenten in der aq_chain und startet das Rendern.
        """
        t1 = time.time()
        request = self.REQUEST
        docs = self.search_xsd_xsi()
        if docs is None:
            return ''
        (xsi, xsd) = docs

        self.check_xsd(xsd)
        
        # fr self.get_session
        request.set('node', xsd.documentElement)

        # some speed improvements
        # those are called a thousend times...
        request.set('xsi_doc_physicalpath', xsi.getPhysicalPath())
        request.set('xsd_doc_physicalpath', xsd.getPhysicalPath())

        xsi_node = xsi.documentElement
        xsi_parent_node = None
        
        xsi_boundary_node_path = self.get_session().get_boundary_node()
        if xsi_boundary_node_path is None:
            xsi_boundary_node = None
        else:
            xsi_boundary_node = xsi.resolveNodePath(xsi_boundary_node_path)

        service_editor = self.service_editor
        service_editor.setDocumentEditor(xsd.documentElement, 'service_schema_normal')

        request.set('xsi_node', xsi_node)
        request.set('xsi_parent_node', xsi_parent_node)
        request.set('xsi_boundary_node', xsi_boundary_node)

        # there's always one element in here. no checking within element/render
        request.set('schema_editor_choice_break', ['foo'])

        t2 = time.time()
        result = service_editor.render(xsd.documentElement)
        t3 = time.time()
        if self.debug_mode(request, 'show_time'):
            request.set('schema_editor_benchmark_rendering', '%.3f sec' % (t3 - t2))
            request.set('schema_editor_benchmark_initializing', '%.3f sec' % (t2 - t1))
        return result

    def set_debug_mode(self, REQUEST, redirect=1):
        """sets or unsets debug_mode"""
        if REQUEST.has_key('set_debug_mode'):
            pass
        session = REQUEST.SESSION
        session.set('schema_editor_debug', REQUEST.get("debug", None))
        if redirect:
            REQUEST.RESPONSE.redirect(self.get_redirect_url())
        return

    def debug_mode(self, REQUEST=None, key=None):
        """searches for debug configuration in SESSION"""
        if REQUEST is None:
            session = self.REQUEST.SESSION
        else:
            session = REQUEST.SESSION
        debug = session.get('schema_editor_debug', None)
        if debug is None:
            return 0
        if key is None:
            return 1
        if debug.has_key(key):
            return 1
        return 0

    def redirect_to_setXmlDocsForm(self, REQUEST):
        """redirects to the setXmlDocsForm."""
        dir = None
        docs = self.search_xsd_xsi(redirect=0)
        if docs is not None:
            (xsi, xsd) = docs
            dir = xsd.getPhysicalPath()[:-1]

        myself = self.getPhysicalPath() + ('setXmlDocsForm',)
        if dir:
            path = '/'.join(minimize_paths(dir, myself))
        else:
            path = '/'.join(myself)
        REQUEST.RESPONSE.redirect(path)
        return

    def setXmlDocs(self, REQUEST):
        """brings the xsi and xsd from the corresponding form into the url"""
        build_from_scratch = 0
        output = ''

        self.set_debug_mode(REQUEST, redirect=0)
        xsi = None
        xsd = None

        if REQUEST.has_key('cleanup'):
            self.cleanup(REQUEST, redirect=0)
        
        context = self.aq_parent
        if REQUEST.has_key('xsd'):
            xsd = context.aq_acquire(REQUEST['xsd'])
        if REQUEST.get('existing_or_new', 'existing') == 'existing':
            if REQUEST.has_key('xsi'):
                xsi = context.aq_acquire(REQUEST['xsi'])
            if REQUEST.get('rebuild_xsi_from_scratch', 'don\'t') == 'rebuild':
                build_from_scratch = 1
        else: # create new XSI
            new_id = REQUEST.get('new_xsi_id', 'new_xsi')
            output += "creating new ParsedXML in %s\n" % context.getId()
            context.manage_addProduct['ParsedXML'].manage_addParsedXML(new_id)
            xsi = context.aq_acquire(new_id)
            output += "xsi created:%s\n" % xsi.getId()
            build_from_scratch = 1

        if build_from_scratch:
            self.check_xsd(xsd)
            output += "create DocumentFragment\n"
            df = xsi.createDocumentFragment()
            REQUEST.set('xsi_parent_node', df)
            REQUEST.set('xsi_node', None)
            output += "calling service_editor\n"
            self.service_editor.setDocumentEditor(xsd.documentElement, 'service_schema_add')
            output += self.service_editor.render(xsd.documentElement) + '\n'
            xsi.replaceChild(df.childNodes[0], xsi.documentElement)
            self.set_edit_node(xsi.documentElement)

        path = '/'.join(minimize_paths(xsi.getPhysicalPath(),
                                       xsd.getPhysicalPath(),
                                       self.getPhysicalPath()))
        if self.debug_mode(REQUEST, 'show_steps') and output:
            return self.result_page(result=output, continue_url=path)
        else:
            REQUEST.RESPONSE.redirect(path)
        return

    def check_xsd(self, xsd):
        """check whether xsd is a schema"""
        schema = xsd.documentElement
        if (schema.namespaceURI != xsdURI or
            schema.localName != 'schema'):
            raise XSExceptions.XSNoSchema('XML Document %s is not an XML Schema Document' %
                                          '/'.join(xsd.getPhysicalPath()))
        if schema.prefix != 'xsd':
            raise XSExceptions.XSNotSupported('The XML Schema namespace has to be bound to the "xsd" prefix. Sorry.')
        return

    def find_xmldocuments(self):
        """returns all 'siblings' of meta_type 'Parsed XML'"""
        return [doc.getId() for doc in self.aq_parent.objectValues('Parsed XML')]

    def get_session(self):
        request = self.REQUEST
        # cache the SchemaEditorSession in REQUEST once per rendering process
        sessiondata = request.get('schema_editor_session_shortcut', None)
        if sessiondata is not None:
            # sessiondata.check(xsd) is not needed.
            # XSD should not change during one rendering prodess...
            return sessiondata
        session = request.SESSION
        if (request.has_key('node') and request.node is not None):
            xsd = request.node.ownerDocument
        else:
            docs = self.search_xsd_xsi()
            if docs is None:
                return
            (xsi, xsd) = docs
        path = request.get('xsd_doc_physicalpath', None)
        if path is None:
            path = xsd.getPhysicalPath()
        
        if not session.has_key('schema_editor_session_data'):
            session['schema_editor_session_data'] = {}
        sessiondict = session['schema_editor_session_data']

        if sessiondict.has_key(path):
            sessiondata = sessiondict[path]
            sessiondata.check(xsd)
        else:
            sessiondata = SchemaEditorSession(xsd, self)
            sessiondict[path] = sessiondata
            sessiondata.check(xsd)
        request.set('schema_editor_session_shortcut', sessiondata)
        return sessiondata

    def set_edit_node(self, node, xsi_boundary_node=None):
        self.get_session().set_edit_node(node, xsi_boundary_node)

    def is_edit_node(self, node, nodepath=None, check_ancestors=0):
        return self.get_session().is_edit_node(node, nodepath, check_ancestors)

    def clear_edit_node(self):
        self.get_session().clear_edit_node()

    def get_element_type(self, element):
        type_name = None
        type_path = None
        # nach einem complexType oder simpleType suchen
        # zuerst unnamed types, dann nach type-attribute suchen
        # siehe XML&Co S 427
        for child in element.childNodes:
            if (child.nodeType == child.ELEMENT_NODE and
                child.namespaceURI == xsdURI and
                child.localName in ('complexType', 'simpleType')):
                type_path = child.getNodePath('element_id')
                break
        else:
            if element.hasAttribute('type'):
                type_name = element.getAttribute('type')
            else:
                # XXX sollte anyType sein (der ist aber noch nicht implementiert)
                type_name = 'xsd:string'
        type_registry = self.get_session().type_registry
        type = type_registry.get_type(name=type_name, nodepath=type_path)
        if type is None:
            raise XSExceptions.XSUnknownTypeException("unknown type %s" % (type_name))
        return type

    def get_type_by_node(self, xsd_node):
        """get_type"""
        path = xsd_node.getNodePath('element_id')
        type_registry = self.get_session().type_registry
        return type_registry.get_type(nodepath=path)

    def get_type_by_name(self, name):
        """searches the type by its name"""
        type_registry = self.get_session().type_registry
        return type_registry.get_type(name=name)

    def get_element_typenode(self, element):
        type = self.get_element_type(element)
        if getattr(type, 'nodepath', None) is not None:
            return element.documentElement.resolveNodePath(type.nodepath)
        return None

    def getCardinality(self, xsd_node, attribute, namespaceURI=None):
        request = self.REQUEST
        # is the cardinality to be overridden?
        # XXX is this too much magic from a widgets point of view??
        if (attribute == 'minOccurs' and
            request.get('schema_editor_override_minOccurs', None) is not None):
            # this is used to add only one instance on an element.
            cardinality = XSCardinality(request.schema_editor_override_minOccurs)
            # all following elements will be created according to the real value of their minOccurs attribute.
            # so "delete" the override value. Sorry, there's no __delitem__ in REQUEST.
            request.set('schema_editor_override_minOccurs', None)
            return cardinality
        
        if xsd_node is None:
            return XSCardinality(attribute)
        else:
            if namespaceURI is None:
                raw_value = xsd_node.getAttribute(attribute)
            else:
                raw_value = xsd_node.getAttributeNS(namespaceURI,
                                                    attribute)
            return XSCardinality(raw_value)

    def pretty_print(self, node):
        result = html_quote(TheXSPrettyPrinter.pprint(node, maxLines=20))
        return result

    def get_redirect_url(self):
        """
        baut die redirect URL auf.
        sucht die XML Dokumente zuerst in REQUEST und dann in der aktuellen URL.
        """
        request = self.REQUEST
        xsi_node = request.get('xsi_node', None)
        if xsi_node is None:
            xsi_node = request.get('xsi_parent_node', None)
        if (request.has_key('node') and
            request.node is not None and xsi_node is not None):
            xsd = request.node.ownerDocument
            xsi = xsi_node.ownerDocument
        else:
            docs = self.search_xsd_xsi()
            if docs is None:
                return ''
            (xsi, xsd) = docs
        path = minimize_paths(xsi.getPhysicalPath(),
                              xsd.getPhysicalPath(),
                              self.getPhysicalPath())
        return '/'.join(path)

    def get_cleanup_url(self):
        #return self.get_redirect_url() + '/cleanup'
        return 'cleanup'

    def cleanup(self, REQUEST, redirect=1):
        """
        does a full cleanup of your XMLWidgets and SchemaEditor session data.
        """
        session = REQUEST.SESSION
        # killing your XMLWidgets Editor Session
        if session.has_key('xmlwidgets_service_editor'):
            session.delete('xmlwidgets_service_editor')
        # killing your SchemaEditor Session
        if session.has_key('schema_editor_session_data'):
            session.delete('schema_editor_session_data')
        if redirect:
            REQUEST.RESPONSE.redirect(REQUEST.URL1)
        return

    def register_type(self, type):
        self.get_session().type_registry.register_type(type)
        return

    def get_image_link(self, url, image_name, alt_text):
        """
        creates a html link to url with an image as links content.
        the image has to be located within the globals folder.
        """
        return self.get_session().get_image_link(url, image_name, alt_text)

manage_addSchemaEditorForm = PageTemplateFile('www/manage_addSchemaEditorForm', globals())

def manage_addSchemaEditor(self, id, REQUEST=None):
    """lala"""
    if REQUEST is not None:
        if REQUEST.get('create_new_folder', None):
            name = REQUEST.get('new_folder_name', None)
            if name:
                self.manage_addFolder(name)
                folder = getattr(self, name)
            else:
                folder = self
        else:
            folder = self
            
        if REQUEST.get('create_helper_objects', None):
            install.install_helpers(folder)
            
        if REQUEST.get('create_demo_documents', None):
            install.install_demo(folder)
    else:
        folder = self
            
    folder._setObject(id, SchemaEditor(id))
    
    if REQUEST is not None:
        return self.manage_main(self, REQUEST)
