# $Header: /cvsroot/SchemaEditor/XSFacet.py,v 1.18 2003/09/05 15:16:24 spuelrich Exp $
__version__ = '$Revision: 1.18 $'[11:-2]

from XSExceptions import XSFacetValidationError, XSNotSupported, XSUnknownFacetException
from XSField import XSField_enumeration
from xsdTools import getElementValue
from Products.PythonScripts.standard import html_quote
import re

class XSFacet:
    def __init__(self, values=None):
        self.values = []
        if values is not None:
            self.values.extend(values)
            
    def add(self, value):
        self.values.append(value)
        
    def validate(self, value, normalized_value):
        pass

# the enumeration facet is a exception to all the other facets.
# it has its own XSField_enumeration to render its values as a ListField
class XSFacet_enumeration(XSFacet):
    name = 'enumeration'
    value_type = None
    field = XSField_enumeration(XSField_enumeration.meta_type, size=1, unicode="1")
    def validate(self, value, normalized_value):
        if not (value in self.values):
            raise XSFacetValidationError('enumeration facet failed: %s not in %s' %
                                         (str(value), str(map(str, level))))
        return value
    
    def render_field(self, xsi_node, value=None):
        self.field.values['items'] = self._get_field_items()
        value = getElementValue(xsi_node)
        value = html_quote(value)
        return self.field.render(xsi_node, value)

    def validate_field(self, xsi_node, REQUEST):
        self.field.values['items'] = self._get_field_items()
        return self.field.validate(xsi_node, REQUEST)

    def _get_field_items(self):
        # converts [1,2,3] to [('1', '1'), ('2', '2'), ('3', '3')]
        result = []
        for facetvalue in self.values:
            s = html_quote(facetvalue.unicode())
            result.append((s, s))
        return result

class XSFacet_maxInclusive(XSFacet):
    name = 'maxInclusive'
    value_type = None # same as the value to validate
    def validate(self, value, normalized_value):
        for facetvalue in self.values:
            if not value <= facetvalue:
                raise XSFacetValidationError('maxInclusive facet failed: %s > %s' %
                                             ((str(value), str(facetvalue))))
        return value
    
class XSFacet_minInclusive(XSFacet):
    name = 'minInclusive'
    value_type = None
    def validate(self, value, normalized_value):
        for facetvalue in self.values:
            if not value >= facetvalue:
                raise XSFacetValidationError('minInclusive facet failed: %s < %s' %
                                             ((str(value), str(facetvalue))))
        return value

class XSFacet_maxExclusive(XSFacet):
    name = 'maxExclusive'
    value_type = None
    def validate(self, value, normalized_value):
        for facetvalue in self.values:
            if not value < facetvalue:
                raise XSFacetValidationError('maxExclusive facet failed: %s >= %s' %
                                             ((str(value), str(facetvalue))))
        return value
    
class XSFacet_minExclusive(XSFacet):
    name = 'minInclusive'
    value_type = None
    def validate(self, value, normalized_value):
        for facetvalue in self.values:
            if not value > facetvalue:
                raise XSFacetValidationError('minInclusive facet failed: %s <= %s' %
                                             ((str(value), str(facetvalue))))
        return value

class XSFacet_length(XSFacet):
    name = 'length'
    value_type = 'nonNegativeInteger'
    def validate(self, value, normalized_value):
        length = len(normalized_value)
        for facetvalue in self.values:
            if length != facetvalue:
                # XXX characters?
                raise XSFacetValidationError('length facet failed: "%s" is not %s elements long' %
                                             (str(normalized_value), str(facetvalue)))
        return value

class XSFacet_minLength(XSFacet):
    name = 'minLength'
    value_type = 'nonNegativeInteger'
    def validate(self, value, normalized_value):
        length = len(normalized_value)
        for facetvalue in self.values:
            if length < facetvalue:
                raise XSFacetValidationError('minLength facet failed: "%s" is too short (%d elements; minimum is %s)' %
                                             (str(normalized_value), length, str(facetvalue)))
        return value

class XSFacet_maxLength(XSFacet):
    name = 'maxLength'
    value_type = 'nonNegativeInteger'
    def validate(self, value, normalized_value):
        length = len(normalized_value)
        for facetvalue in self.values:
            if length > facetvalue:
                raise XSFacetValidationError('maxLength facet failed: "%s" is too long (%d elements; only %s allowed).' %
                                             (str(normalized_value), length, str(facetvalue)))
        return value

class XSFacet_pattern(XSFacet):
    name = 'pattern'
    value_type = 'string'
    def add(self, value):
        # re.compile doesn't like XSType_nofacet_string
        value = str(value)
        self.values.append(re.compile(value))
        return
    
    def validate(self, value, normalized_value):
        # the value has to match only one pattern per level (OR),
        # but every level must pass (AND).
        for pattern in self.values:
            result = pattern.match(normalized_value)
            # test, whether the pattern matched the whole string; not only the beginning
            if result is not None and result.end() == len(normalized_value):
                # this level passed; ignore the rest
                break
        else:
            # no pattern within this level matched. error!
            patterns = []
            for pattern in self.values:
                patterns.append(pattern.pattern)
            errstr = u'pattern facet failed: "%s" does not match any in %s' % (normalized_value,
                                                                               str(patterns))
            raise XSFacetValidationError(errstr)
        return value

class XSFacet_whiteSpace(XSFacet):
    name = 'whiteSpace'
    value_type = 'string'
    def validate(self, value, normalized_value):
        facetvalue = self.values[-1]
        if facetvalue == 'preserve':
            return value
        elif facetvalue == 'replace':
            return re.sub('\s', ' ', value)
        elif facetvalue == 'collapse':
            return re.sub('\s+', ' ', value)
        else:
            raise XSNotSupported('unsupported value for whitespace facet: "%s"' % facetvalue)

class XSFacet_totalDigits(XSFacet):
    name = 'totalDigits'
    value_type = 'positiveInteger'
    def validate(self, value, normalized_value):
        # find all digits in the string representation of value
        #count = len(re.findall('\d', str(value)))
        count = len(re.findall('\d', normalized_value))
        for facetvalue in self.values:
            if count > facetvalue:
                raise XSFacetValidationError('totalDigits facet failed: "%s" has more than %s digits' %
                                             (str(value), str(facetvalue)))
        return value

class XSFacet_fractionDigits(XSFacet):
    name = 'fractionDigits'
    value_type = 'nonNegativeInteger'
    def validate(self, value, normalized_value):
        valuestring = normalized_value
        parts = valuestring.split('.', 1)
        if len(parts) == 1:
            # no fraction digits
            # is always ok
            return value
        # len(parts) > 1
        fraction = parts[-1]
        # find all digits in the fraction
        count = len(re.findall('\d', fraction))
        for facetvalue in self.values:
            if count > facetvalue:
                raise XSFacetValidationError('fractionDigits facet failed: "%s" has more than %s fraction digits' %
                                             (valuestring, str(facetvalue)))
        return value

class XSFacetRegistry:
    def __init__(self):
        self.facets = {}
        
    def register_facet(self, klass):
        self.facets[klass.name] = klass

    def get_facet(self, name, values=None):
        if self.facets.has_key(name):
            return self.facets[name](values)
        else:
            raise XSUnknownFacetException('facet "%s" is unknown' % name)

XSFacetRegistry = XSFacetRegistry()

XSFacetRegistry.register_facet(XSFacet_maxInclusive)
XSFacetRegistry.register_facet(XSFacet_minInclusive)
XSFacetRegistry.register_facet(XSFacet_maxExclusive)
XSFacetRegistry.register_facet(XSFacet_minInclusive)
XSFacetRegistry.register_facet(XSFacet_enumeration)
XSFacetRegistry.register_facet(XSFacet_length)
XSFacetRegistry.register_facet(XSFacet_minLength)
XSFacetRegistry.register_facet(XSFacet_maxLength)
XSFacetRegistry.register_facet(XSFacet_pattern)
XSFacetRegistry.register_facet(XSFacet_whiteSpace)
XSFacetRegistry.register_facet(XSFacet_totalDigits)
XSFacetRegistry.register_facet(XSFacet_fractionDigits)
